"""
Ollama models
"""

import logging

import requests
from requests import Response

from fence.models.base import LLM, get_log_callback

logger = logging.getLogger(__name__)


class OllamaBase(LLM):
    """Base class for Ollama-powered models"""

    inference_type = "ollama"

    def __init__(
        self,
        source: str,
        metric_prefix: str | None = None,
        extra_tags: dict | None = None,
        endpoint: str | None = None,
        **kwargs,
    ):
        """
        Initialize a Claude model

        :param str source: An indicator of where (e.g., which feature) the model is operating from
        :param str|None metric_prefix: Prefix for the metric names
        :param dict|None extra_tags: Additional tags to add to the logging tags
        :param **kwargs: Additional keyword arguments
        """

        super().__init__(
            source=source, metric_prefix=metric_prefix, extra_tags=extra_tags
        )

        # Set default generate endpoint
        self.endpoint = endpoint or "http://localhost:11434/api"
        self.generate_endpoint = self.endpoint + "/generate"
        self.tag_endpoint = self.endpoint + "/tags"
        self.pull_endpoint = self.endpoint + "/pull"

        # Check if the endpoint is valid
        try:
            requests.get(self.tag_endpoint)
        except requests.exceptions.ConnectionError:
            raise ValueError(
                f"No Ollama service found at {self.endpoint}. Try installing it using `brew install ollama`."
            )

        # LLM parameters
        self.model_kwargs = {
            "temperature": kwargs.get("temperature", 0.01),
            "max_tokens_to_sample": kwargs.get("max_tokens_to_sample", 2048),
        }

    def invoke(self, prompt: str, **kwargs) -> str:
        """
        Call the model with the given prompt
        :param prompt: text to feed the model
        :param bare_bones: whether to use the 'Human: ... Assistant: ...' formatting
        :return: response
        """

        # Call the model
        response = self._invoke(prompt=prompt)

        # Get completion
        completion = response["response"]

        # Get input and output tokens
        input_token_count = response["prompt_eval_count"]
        output_token_count = response["eval_count"]

        # Get input and output word count
        input_word_count = len(prompt.split())
        output_word_count = len(completion.split())

        # Log all metrics if a log callback is registered
        if log_callback := get_log_callback():
            prefix = ".".join(
                item for item in [self.metric_prefix, self.source] if item
            )
            log_args = [
                # Add metrics
                {
                    f"{prefix}.invocation": 1,
                    f"{prefix}.input_token_count": input_token_count,
                    f"{prefix}.output_token_count": output_token_count,
                    f"{prefix}.input_word_count": input_word_count,
                    f"{prefix}.output_word_count": output_word_count,
                },
                # Add tags
                self.logging_tags,
            ]

            # Log metrics
            logger.debug(f"Logging args: {log_args}")
            log_callback(*log_args)

        return completion

    def _invoke(self, prompt: str) -> dict[str, str]:
        """
        Handle the API request to the service
        :param prompt: text to feed the model
        :return: response completion
        """

        # Payload for post request to Ollama
        payload = {
            "prompt": prompt,
            "options": self.model_kwargs,
            "model": self.model_id,
            "stream": False,
        }

        # Send request
        try:
            response = requests.post(url=self.generate_endpoint, json=payload)

        except Exception as e:
            raise ValueError(f"Error raised by Ollama service: {e}")

        # Check if the response is valid
        if response.status_code != 200 and response.text.__contains__("not found"):
            logger.warning(
                f"Model {self.model_id} not found in Ollama service - trying to pull it"
            )
            self._pull_model(model_id=self.model_id)

            # Retry the request
            try:
                response = requests.post(url=self.generate_endpoint, json=payload)
            except Exception as e:
                raise ValueError(f"Error raised by Ollama service: {e}")

        return response.json()

    def _pull_model(self, model_id: str) -> Response:
        """
        Pull the model from the Ollama service
        :param model_id: model name
        :return: model
        """

        # Send request
        try:
            response = requests.post(
                url=self.pull_endpoint, json={"name": model_id}, stream=False
            )

        except Exception as e:
            raise ValueError(f"Error raised by Ollama service: {e}")

        return response

    def _get_model_list(self) -> list[str]:
        """
        Get the list of available models from the Ollama service
        :return: list of model names
        """

        # Send request
        try:
            response = requests.get(url=self.tag_endpoint)

        except Exception as e:
            raise ValueError(f"Error raised by Ollama service: {e}")

        return response.json()["models"]


class Ollama(OllamaBase):
    """Generic Ollama interface"""

    def __init__(self, model_id: str, source: str, auto_pull: bool = True, **kwargs):
        """
        Initialize the Ollama model
        :param str model_id: The model ID as defined in the Ollama service
        :param str source: An indicator of where (e.g., which feature) the model is operating from.
        :param bool auto_pull: Whether to automatically pull the model if it is not found
        :param **kwargs: Additional keyword arguments
        """

        super().__init__(source=source, auto_pull=auto_pull, **kwargs)

        self.model_id = model_id
        self.model_name = f"{model_id.capitalize()}"


class Llama3_1(Ollama):
    """Llama 3.1 model"""

    def __init__(self, source: str, **kwargs):
        """
        Initialize the Llama 3.1 model
        :param str source: An indicator of where (e.g., which feature) the model is operating from.
        :param **kwargs: Additional keyword arguments
        """

        super().__init__(model_id="llama3.1", source=source, **kwargs)


if __name__ == "__main__":

    # Initialize the model
    model = Llama3_1(source="test")

    # List available models
    model_list = model._get_model_list()

    # Call the model
    response = model("Hello, how are you?")

    # Try a non-existent model
    model = Ollama(model_id="gemma2", source="test")

    # Call the model
    response_gemma = model("Hello, which model are you?")



"""
Tool using agent
"""

import logging

from fence import LLM, Link, MessagesTemplate
from fence.agents.base import BaseAgent
from fence.links import logger as link_logger
from fence.memory.base import BaseMemory, FleetingMemory
from fence.models.openai import GPT4omini
from fence.prompts.agents import CHAT_PROMPT

logger = logging.getLogger(__name__)


# Suppress the link logger
link_logger.setLevel("CRITICAL")


class ChatAgent(BaseAgent):
    """An LLM-based Agent, designed to chat with a user or another agent"""

    def __init__(
        self,
        identifier: str | None = None,
        model: LLM | None = None,
        description: str | None = None,
        memory: BaseMemory | None = None,
        name: str | None = None,
        profile: str | None = None,
    ):
        """
        Initialize the Agent object.

        :param identifier: An identifier for the agent. If none is provided, the class name will be used.
        :param model: An LLM model object.
        :param description: A description of the agent.
        :param memory: A memory object.
        :param name: The name of the agent. Will be used in multi-agent conversations.
        :param profile: A behavioral profile of the agent.
        """

        super().__init__(identifier=identifier, model=model, description=description)

        logger.info(f"Creating an agent with model: {model.model_name} ")

        # Set the name
        self.name = name or None

        # Create a memory context for the agent
        self.context = (memory or FleetingMemory)()
        self.context.add_message(
            role="system", content=CHAT_PROMPT.format(profile=profile or description)
        )

    def run(self, prompt: str) -> str:
        """Run the agent with the given prompt"""

        # Add the prompt to the history
        self.context.add_message(role="user", content=prompt)

        # Base link
        link = Link(
            name="agent_step",
            model=self.model,
            template=MessagesTemplate(source=self.context),
        )

        response = link.run()["state"]

        # If we have a name, and the model added it to the output itself, remove it
        if self.name and response.startswith(f"{self.name}: "):
            response = response[len(self.name) + 2 :]

        # Log the response
        logger.info(f"{self.name}: {response}")

        # Add the response to the history
        self.context.add_message(
            role="assistant",
            content=f"{f'{self.name}: ' if self.name else ''}{response}",
        )

        return response


if __name__ == "__main__":

    # Set up the agent
    agent = ChatAgent(
        model=GPT4omini(
            source="chat",
        ),
    )

    # Loop to chat
    while True:
        prompt = input("You: ")
        response = agent.run(prompt)
        print(f"Agent: {response}")
        if response == "Goodbye!":
            break



from fence.chains import Chain, LinearChain
from fence.links import Link, TransformationLink
from fence.models.base import LLM
from fence.models.claude import ClaudeBase, ClaudeInstant, ClaudeV2
from fence.models.claude3 import Claude3Base, Claude35Sonnet, ClaudeHaiku, ClaudeSonnet
from fence.models.openai import GPT4o, GPTBase
from fence.parsers import BoolParser, IntParser, TOMLParser, TripleBacktickParser
from fence.templates import Message, Messages, MessagesTemplate
from fence.templates.string import StringTemplate
from fence.utils.logger import setup_logging

__all__ = [
    "Link",
    "TransformationLink",
    "LLM",
    "ClaudeBase",
    "ClaudeInstant",
    "ClaudeV2",
    "Claude3Base",
    "ClaudeHaiku",
    "ClaudeSonnet",
    "Claude35Sonnet",
    "GPTBase",
    "GPT4o",
    "BoolParser",
    "IntParser",
    "TOMLParser",
    "TripleBacktickParser",
    "MessagesTemplate",
    "Message",
    "Messages",
    "StringTemplate",
    "Chain",
    "LinearChain",
    "setup_logging",
]



"""
This module contains data models used in the API formatting.
"""

import base64
from mimetypes import guess_type
from pathlib import Path
from typing import Literal, Optional

from pydantic import BaseModel, model_validator

##########
# Models #
##########


class Source(BaseModel):
    """A model representing a media source. Can be base64 encoded data or a URL."""

    type: Literal["base64"] = "base64"
    media_type: Literal["image/jpeg", "image/png", "image/webp", "image/gif"]
    data: str | None = None
    _file_path: str | Path | None = None

    # Validate that only one of 'data' or 'url' is provided
    @model_validator(mode="before")
    def handle_data_or_url(cls, values):

        # Extract data and file_path values
        data, file_path = values.get("data"), values.get("_file_path")

        # Cannot provide both data and file_path
        if data and file_path:
            raise ValueError("Only one of 'data' or 'file_path' should be provided.")

        # Must provide either data or file_path
        if not data and not file_path:
            raise ValueError("One of 'data' or 'file_path' must be provided.")

        # If a file path is provided, check if it is a valid local path
        if file_path:
            with open(file_path, "rb") as file:
                image_data = file.read()
                values["data"] = base64.b64encode(image_data).decode("utf-8")
                values["media_type"], _ = guess_type(file_path)

        return values


class TextContent(BaseModel):
    """A model representing text content."""

    type: Literal["text"] = "text"
    text: str


class ImageContent(BaseModel):
    """A model representing image content."""

    type: Literal["image"] = "image"
    source: Source


# Content can be either text or image
Content = TextContent | ImageContent


class Message(BaseModel):
    """A base model representing a message."""

    role: Literal["user", "assistant"]
    content: list[Content] | str


class Messages(BaseModel):
    """A model representing a collection of messages."""

    messages: list[Message]
    system: Optional[str] = None



import logging

from .base import BaseTool

logger = logging.getLogger(__name__)


class TextInverterTool(BaseTool):
    """Inverts the text"""

    def execute_tool(self, text: str, **kwargs) -> str:
        """Invert the text"""
        logger.info(f"{self.__class__.__name__} was called: {text}")

        return text[::-1]


# class CalculatorTool(BaseTool):
#     """Perform mathematical calculations"""
#
#     def execute_tool(self, expression: str, **kwargs) -> float | str:
#         """
#         Evaluate the mathematical expression.
#         Optionally, environment can influence behavior, like precision or logging.
#         """
#         logger.info(f"{self.__class__.__name__} was called with expression: {expression}")
#         try:
#             import numexpr
#
#             result = numexpr.evaluate(expression)
#
#             # Access environment variables directly from kwargs
#             environment = kwargs.get('environment', {})
#
#             # Optional precision from environment
#             precision = environment.get('precision', None)
#             if precision is not None and isinstance(result, (int, float)):
#                 result = round(result, precision)
#
#             # Optionally adjust logging based on environment
#             log_level = environment.get('log_level', 'info')
#             if log_level == 'debug':
#                 logger.debug(f"Expression: {expression}, Result: {result}")
#             else:
#                 logger.info(f"Expression: {expression}, Result: {result}")
#
#         except Exception as e:
#             logger.error(f"Error evaluating expression: {e}", exc_info=True)
#             return f"Error evaluating expression {expression} - {e}"
#
#         return result



"""
Claude Gen 3 models
"""

import json
import logging

import boto3

from fence.models.base import (
    LLM,
    MessagesMixin,
    get_log_callback,
    register_log_callback,
    register_log_tags,
)
from fence.templates.models import Messages

MODEL_ID_SONNET = "anthropic.claude-3-sonnet-20240229-v1:0"
MODEL_ID_HAIKU = "anthropic.claude-3-haiku-20240307-v1:0"

logger = logging.getLogger(__name__)


class Claude3Base(LLM, MessagesMixin):
    """Base class for Claude (gen 3) models"""

    inference_type = "bedrock"

    def __init__(
        self,
        source: str | None = None,
        full_response: bool = False,
        metric_prefix: str | None = None,
        extra_tags: dict | None = None,
        **kwargs,
    ):
        """
        Initialize a Claude model

        :param str source: An indicator of where (e.g., which feature) the model is operating from.
        :param bool full_response: whether to return the full response or just the TEXT completion
        :param str|None metric_prefix: Prefix for the metric names
        :param dict|None extra_tags: Additional tags to add to the logging tags
        :param **kwargs: Additional keyword arguments
        """

        super().__init__(metric_prefix=metric_prefix, extra_tags=extra_tags)

        self.full_response = full_response

        # LLM parameters
        self.model_kwargs = {
            "temperature": kwargs.get("temperature", 0.01),
            "max_tokens": kwargs.get("max_tokens", 2048),
        }

        # AWS parameters
        self.region = kwargs.get("region", "eu-central-1")
        self.client = boto3.client("bedrock-runtime", self.region)

    def invoke(self, prompt: str | Messages, **kwargs) -> str:
        """
        Call the model with the given prompt
        :param prompt: text to feed the model
        :return: response
        """

        # Call the model
        response = self._invoke(prompt=prompt)

        # Get response completion
        response_body = json.loads(response.get("body").read().decode())

        # Get token counts
        input_token_count = response_body["usage"]["input_tokens"]
        output_token_count = response_body["usage"]["output_tokens"]

        # Get completion
        completion = response_body["content"][0]["text"]

        # Get input and output word count
        if isinstance(prompt, str):
            input_word_count = len(prompt.split())
        elif isinstance(prompt, Messages):
            input_word_count = self.count_words_in_messages(prompt)
        else:
            raise ValueError(
                f"Prompt must be a string or a list of messages. Got {type(prompt)}"
            )
        output_word_count = len(completion.split())

        # Log all metrics if a log callback is registered
        if log_callback := get_log_callback():
            prefix = ".".join(
                item for item in [self.metric_prefix, self.source] if item
            )
            log_args = [
                # Add metrics
                {
                    f"{prefix}.invocation": 1,
                    f"{prefix}.input_token_count": input_token_count,
                    f"{prefix}.output_token_count": output_token_count,
                    f"{prefix}.input_word_count": input_word_count,
                    f"{prefix}.output_word_count": output_word_count,
                },
                # Add tags
                self.logging_tags,
            ]

            # Log metrics
            logger.debug(f"Logging args: {log_args}")
            log_callback(*log_args)

        return completion

    def _invoke(self, prompt: str | Messages) -> dict:
        """
        Handle the API request to the service
        :param prompt: text to feed the model
        :return: response completion
        """

        # Format prompt: Claude3 models expect a list of messages, with content, roles, etc.
        # However, if we receive a string, we will format it as a single user message for ease of use.
        if isinstance(prompt, Messages):
            messages = prompt.model_dump(exclude_none=True)
        elif isinstance(prompt, str):
            messages = {
                "messages": [
                    {
                        "role": "user",
                        "content": [
                            {
                                "type": "text",
                                "text": prompt,
                            }
                        ],
                    }
                ]
            }
        else:
            raise ValueError("Prompt must be a string or a list of messages")

        # Build request body
        request_body = {
            "anthropic_version": "bedrock-2023-05-31",
            **messages,
            **self.model_kwargs,
        }

        logger.debug(f"Request body: {request_body}")

        # Send request
        try:
            response = self.client.invoke_model(
                modelId=self.model_id,
                contentType="application/json",
                accept="application/json",
                body=json.dumps(request_body),
            )

            # Log invocation
            self.invocation_logging()

            return response

        except Exception as e:
            raise ValueError(f"Error raised by bedrock service: {e}")

    def invocation_logging(self) -> None:
        """
        Log invocation
        :return:
        """
        pass


class ClaudeHaiku(Claude3Base):
    """Claude Haiku model class"""

    def __init__(self, **kwargs):
        """
        Initialize a Claude Haiku model
        :param str source: An indicator of where (e.g., which feature) the model is operating from.
        :param **kwargs: Additional keyword arguments
        """

        super().__init__(**kwargs)

        self.model_id = MODEL_ID_HAIKU
        self.model_name = "ClaudeHaiku"


class ClaudeSonnet(Claude3Base):
    """Claude Sonnet model class"""

    def __init__(self, **kwargs):
        """
        Initialize a Claude Sonnet model
        :param str source: An indicator of where (e.g., which feature) the model is operating from.
        :param **kwargs: Additional keyword arguments
        """

        super().__init__(**kwargs)

        self.model_id = MODEL_ID_SONNET
        self.model_name = "ClaudeSonnet"


class Claude35Sonnet(Claude3Base):
    """Claude Sonnet 3.5 model class"""

    def __init__(self, **kwargs):
        """
        Initialize a Claude Sonnet 3.5 model
        :param str source: An indicator of where (e.g., which feature) the model is operating from.
        :param **kwargs: Additional keyword arguments
        """

        super().__init__(**kwargs)

        self.model_id = "anthropic.claude-3-5-sonnet-20240620-v1:0"
        self.model_name = "ClaudeSonnet3.5"


if __name__ == "__main__":

    # Register logging callback
    register_log_callback(lambda metrics, tags: print(metrics, tags))

    # Register logging tags
    register_log_tags({"team": "data-science-test", "project": "fence"})

    # Create an instance of the ClaudeHaiku class
    claude_sonnet = Claude35Sonnet(
        source="test", metric_prefix="yolo", region="us-east-1"
    )

    # Call the invoke method with a prompt
    response = claude_sonnet.invoke(prompt="The sun is shining brightly")



"""
Agent class to orchestrate a flow that potentially calls tools or other, specialized agents.
"""

import logging

from fence import LLM, Link, MessagesTemplate, TOMLParser
from fence.agents.base import BaseAgent
from fence.links import logger as link_logger
from fence.memory.base import BaseMemory, FleetingMemory
from fence.prompts.agents import REACT_MULTI_AGENT_PROMPT

logger = logging.getLogger(__name__)

# Suppress the link logger
link_logger.setLevel("INFO")


class RefereeAgent(BaseAgent):
    """An LLM-based Agent, capable of delegating to other agents"""

    def __init__(
        self,
        model: LLM = None,
        description: str | None = None,
        delegates: list[BaseAgent] = None,
        memory: BaseMemory = None,
    ):
        """
        Initialize the Referee agent object.

        :param model: An LLM model object.
        :param delegates: A list of Agent objects, to which the Referee can delegate.
        :param memory: A memory class.
        """

        super().__init__(model=model, description=description)

        # Store the tools and their names
        self.delegates = delegates or []
        self.delegate_names = [agent.__class__.__name__ for agent in self.delegates]

        logger.info(
            f"Creating an agent with model <{model.model_name}> and delegates: {self.delegate_names}"
        )

        # Create a memory context for the agent
        self.memory = memory
        self.context = None
        self.wipe_context(memory=self.memory)

        # Format tools for the prompt
        self.formatted_agents = "".join(agent.format_toml() for agent in self.delegates)
        logger.debug(f"Tools: {self.formatted_agents}")

    def run(self, prompt: str) -> str:
        """Run the agent with the given prompt"""

        # Add the prompt to the history
        self.context.add_message(role="user", content=prompt)

        ################
        # Agentic loop #
        ################
        response = None
        answer = None
        iteration_count = 0
        max_iterations = 30  # safeguard against infinite loops

        while response is None or not response.startswith("[ANSWER]"):

            # Increment iteration counter and check for loop limit
            iteration_count += 1
            if iteration_count > max_iterations:
                logger.warning("Max iterations reached, exiting loop.")
                break

            # Base link
            link = Link(
                name="referee_step",
                model=self.model,
                template=MessagesTemplate(source=self.context),
            )

            response = link.run(input_dict={"agents": self.formatted_agents})["state"]
            logger.info(f"Model response: {response}")

            # Add the response to the history
            self.context.add_message(role="assistant", content=response)

            # If the response contains an answer, break the loop
            if "[ANSWER]" in response:
                answer = response.split("[ANSWER]")[-1].strip()
                break

            # Check if the response contains a tool action
            delegate_response = None
            if "[DELEGATE]" in response:

                logger.debug("Found an agent in the response")

                # Extract the action from the response
                delegate = response.split("[DELEGATE]")[-1].strip()
                logger.debug(f"Delegate: {delegate}")

                # Extract the tool name and parameters
                parser = TOMLParser()
                try:
                    delegate_parsed = parser.parse(input_string=delegate)
                    logger.debug(f"Delegate Parsed: {delegate_parsed}")
                except Exception as e:
                    logger.error(f"Failed to parse delegate agent: {e}")
                    prompt = "Invalid delegate"
                    self.context.add_message(
                        role="assistant", content="Invalid delegate"
                    )
                    continue

                # If the action is valid, run the tool
                if (
                    delegate_parsed
                    and "agent_name" in delegate_parsed
                    and "agent_input" in delegate_parsed
                ):
                    delegate_name = delegate_parsed["agent_name"]
                    delegate_input = delegate_parsed["agent_input"]
                    logger.info(
                        f"Executing tool: {delegate_name} with input: {delegate_input}"
                    )

                    # Check if the tool exists
                    if delegate_name in self.delegate_names:
                        logger.debug(
                            f"Running tool: {delegate_name} with input: {delegate_input}"
                        )
                        delegate = self.delegates[
                            self.delegate_names.index(delegate_name)
                        ]
                        try:
                            delegate_response = delegate.run(prompt=delegate_input)
                            logger.info(f"Delegate Response: {delegate_response}")
                        except Exception as e:
                            logger.error(f"Error running delegate {delegate_name}: {e}")
                            delegate_response = "Delegate execution failed"
                    else:
                        logger.error(f"Delegate {delegate_name} not found")
                        delegate_response = "Delegate not found"
                        continue
                else:
                    prompt = "Invalid delegate"
                    self.context.add_message(
                        role="assistant", content="Invalid delegate"
                    )
                    continue

            # Add the response to the history
            self.context.add_message(
                role="user",
                content=(
                    f"[OBSERVATION] {delegate_response}"
                    if delegate_response is not None
                    else "No delegate response"
                ),
            )

        # Clear the memory context
        self.wipe_context(self.memory)

        return answer

    def wipe_context(self, memory: BaseMemory | None = None):
        """Clear the memory context"""
        self.context = (
            memory if isinstance(memory, BaseMemory) else (memory or FleetingMemory)()
        )
        self.context.add_message(role="system", content=REACT_MULTI_AGENT_PROMPT)



import logging
import re
from abc import ABC, abstractmethod

from fence.templates.models import Messages

logger = logging.getLogger(__name__)


class BaseTemplate(ABC):
    """
    Base class for template objects.
    """

    def __init__(self, source: str | Messages, **kwargs):
        """
        Initialize the BaseTemplate object.

        :param str | Messages source: The template string, or a list of Message objects.
        :param list input_variables: List of input variables expected by the template.
        :param kwargs: Additional keyword arguments.
            - separator (str, optional): Separator to use when merging string-based templates. Default is a space.
        """

        self.source = source
        self.type = type(source)
        self.input_variables = []

        # We have no time for anything other than strings or Messages
        if not isinstance(source, (str, Messages)):
            raise TypeError(
                f"Template must be a string or a Messages object - received {type(source)}"
            )

    def __call__(self, input_dict: dict = None, **kwargs):
        """
        Render the template with the provided variables.

        :param kwargs: Keyword arguments containing values for the template variables.
        :return: Rendered template string.
        :rtype: str

        :raises ValueError: If any required variable is missing.
        """
        return self.render(input_dict=input_dict, **kwargs)

    @abstractmethod
    def render(self, input_dict: dict = None, **kwargs):
        """
        Render the template using the input variables.

        :param input_dict: Dictionary of input variables.
        :param kwargs: Input variables for the template.
        :return: The rendered template.
        """
        pass

    @abstractmethod
    def _find_placeholders(self) -> list[str]:
        """
        Find placeholders in a list of messages. Placeholders are defined as text enclosed in double curly braces.
        They are used to denote variables that need to be replaced in the messages.
        """
        pass

    def _validate_input(self, input_dict: dict):
        # Find both missing and superfluous variables
        missing_variables = [
            variable for variable in self.input_variables if variable not in input_dict
        ]
        superfluous_variables = [
            variable for variable in input_dict if variable not in self.input_variables
        ]
        if missing_variables:
            raise ValueError(f"Missing variables: {missing_variables}")
        if superfluous_variables:
            logger.debug(f"Superfluous variables: {superfluous_variables}")

    ###############
    # Class utils #
    ###############

    def _find_placeholders_in_text(self, text: str) -> list[str]:
        """
        Find placeholders in a text.

        :param str text: The text.
        :return: List of placeholders
        :rtype: list[str]
        """

        return re.findall(r"\{\s*([a-zA-Z0-9_]+)\s*}", text)

    def _render_string(self, text: str, input_dict: dict = None, **kwargs):
        """
        Render a string template with the provided variables.

        :param str text: The template string.
        :param dict input_dict: Dictionary of input variables.
        :param kwargs: Keyword arguments containing values for the template variables.
        :return: Rendered string.
        :raises ValueError: If any required variable is missing.
        """

        # Make input_dict
        if input_dict is None:
            input_dict = {}

        input_dict.update(kwargs)

        # Find both missing and superfluous variables
        self._validate_input(input_dict=input_dict)

        return text.format(**input_dict)

    def __str__(self):
        return f"{self.__class__}: {self.source}"



"""
Claude Gen 1/2 models
"""

import json
import logging

import boto3

from fence.models.base import (
    LLM,
    get_log_callback,
    register_log_callback,
    register_log_tags,
)

logger = logging.getLogger(__name__)


class ClaudeBase(LLM):
    """Base class for Claude (gen 1/2) models"""

    inference_type = "bedrock"

    def __init__(
        self,
        source: str | None = None,
        metric_prefix: str | None = None,
        extra_tags: dict | None = None,
        **kwargs,
    ):
        """
        Initialize a Claude model

        :param str source: An indicator of where (e.g., which feature) the model is operating from
        :param str|None metric_prefix: Prefix for the metric names
        :param dict|None extra_tags: Additional tags to add to the logging tags
        :param **kwargs: Additional keyword arguments
        """

        super().__init__(
            source=source, metric_prefix=metric_prefix, extra_tags=extra_tags
        )

        # LLM parameters
        self.model_kwargs = {
            "temperature": kwargs.get("temperature", 0.01),
            "max_tokens_to_sample": kwargs.get("max_tokens_to_sample", 2048),
        }

        # AWS parameters
        self.region = kwargs.get("region", "eu-central-1")
        self.client = boto3.client("bedrock-runtime", self.region)

    def invoke(self, prompt: str, **kwargs) -> str:
        """
        Call the model with the given prompt
        :param prompt: text to feed the model
        :param bare_bones: whether to use the 'Human: ... Assistant: ...' formatting
        :return: response
        """

        # Format response
        claude_prompt = f"\n\nHuman: {prompt}\n\nAssistant: "

        # Call the model
        response = self._invoke(prompt=claude_prompt)

        # Get input and output tokens
        input_token_count = response["ResponseMetadata"]["HTTPHeaders"][
            "x-amzn-bedrock-input-token-count"
        ]
        output_token_count = response["ResponseMetadata"]["HTTPHeaders"][
            "x-amzn-bedrock-output-token-count"
        ]

        # Get response completion
        response_body = json.loads(response.get("body").read().decode())
        completion = response_body.get("completion")

        # Get input and output word count
        input_word_count = len(prompt.split())
        output_word_count = len(completion.split())

        # Log all metrics if a log callback is registered
        if log_callback := get_log_callback():
            prefix = ".".join(
                item for item in [self.metric_prefix, self.source] if item
            )
            log_args = [
                # Add metrics
                {
                    f"{prefix}.invocation": 1,
                    f"{prefix}.input_token_count": input_token_count,
                    f"{prefix}.output_token_count": output_token_count,
                    f"{prefix}.input_word_count": input_word_count,
                    f"{prefix}.output_word_count": output_word_count,
                },
                # Add tags
                self.logging_tags,
            ]

            # Log metrics
            logger.debug(f"Logging args: {log_args}")
            log_callback(*log_args)

        return completion

    def _invoke(self, prompt: str) -> str:
        """
        Handle the API request to the service
        :param prompt: text to feed the model
        :return: response completion
        """

        # Payload for post request to bedrock
        input_body = {**self.model_kwargs, "prompt": prompt}
        body = json.dumps(input_body)

        # Headers
        accept = "application/json"
        contentType = "application/json"

        # Send request
        try:
            response = self.client.invoke_model(
                body=body,
                modelId=self.model_id,
                accept=accept,
                contentType=contentType,
            )

            return response

        except Exception as e:
            raise ValueError(f"Error raised by bedrock service: {e}")


class ClaudeInstant(ClaudeBase):
    """Claude Instant model class"""

    def __init__(self, **kwargs):
        """
        Initialize a Claude Instant model
        :param str source: An indicator of where (e.g., which feature) the model is operating from.
        :param **kwargs: Additional keyword arguments
        """

        super().__init__(**kwargs)

        self.model_id = "anthropic.claude-instant-v1"
        self.model_name = "ClaudeInstant"


class ClaudeV2(ClaudeBase):
    """Claude v2 model class"""

    def __init__(self, **kwargs):
        """
        Initialize a Claude Instant model
        :param str source: An indicator of where (e.g., which feature) the model is operating from.
        :param **kwargs: Additional keyword arguments
        """

        super().__init__(**kwargs)

        self.model_id = "anthropic.anthropic.claude-v2"
        self.model_name = "ClaudeV2"


if __name__ == "__main__":

    # Register logging callback
    register_log_callback(lambda metrics, tags: print(metrics, tags))

    # Register logging tags
    register_log_tags({"team": "data-science-test", "project": "fence"})

    # Initialize Claude Instant model
    claude = ClaudeInstant(
        source="test", metric_prefix="supertest", extra_tags={"test": "test"}
    )

    # Call the model
    response = claude("Hello, how are you?")



"""
Various utils
"""

import functools
import logging
import time
from concurrent.futures import ThreadPoolExecutor, TimeoutError
from pathlib import Path

CONF_DIR = Path(__file__).resolve().parent.parent.parent / "conf"
DATA_DIR = Path(__file__).resolve().parent.parent.parent / "data"

logger = logging.getLogger(__name__)


def time_it(f=None, threshold: int = 300, only_warn: bool = True):
    """
    Timer decorator: shows how long execution of function took.
    :param f: function to measure
    :param threshold: threshold for warning
    :param only_warn: whether to only log a warning if the threshold is exceeded
    """
    # Check if the decorator is used without parentheses
    if f is None:
        return lambda func: time_it(func, threshold=threshold, only_warn=only_warn)

    def timed(*args, **kwargs):
        t1 = time.time()
        res = f(*args, **kwargs)
        t2 = time.time()
        duration = round(t2 - t1, 2)

        # If the first argument is an object, get its class name
        class_name, object_name = None, None
        if args:
            class_name = (
                args[0].__class__.__name__
                if hasattr(args[0], "__class__") and not isinstance(args[0], dict)
                else None
            )
            object_name = args[0].name if hasattr(args[0], "name") else None
        method_name = f.__name__
        message = f"Method <{class_name if class_name else ''}{(f'[{object_name}]' if object_name else '')}{'.' if any([class_name, object_name]) else ''}{method_name}> took {duration} s to execute."
        if duration > threshold:
            logger.warning(message)
        elif not only_warn:
            logger.info(message)
        else:
            pass

        return res

    return timed


def time_out(f=None, seconds: int = 60):
    """Decorator that adds a timeout to a function.

    :param f: Function to be executed.
    :param seconds: Timeout in seconds.
    :return: Wrapped function with a timeout.
    """

    if f is None:
        return lambda func: time_out(func, seconds=seconds)

    @functools.wraps(f)
    def wrapper_timeout(*args, **kwargs):
        """Wrapper function for the timeout decorator."""

        def run_func(*args, **kwargs):
            """Execute the original function and handle exceptions."""
            try:
                return f(*args, **kwargs)
            except Exception as e:
                raise e

        # Use a ThreadPoolExecutor to run the function in a separate thread
        with ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(run_func, *args, **kwargs)

            try:
                return future.result(timeout=seconds)
            except TimeoutError:
                raise TimeoutError(f"Function call <{f.__name__}> timed out")

    return wrapper_timeout


def setup_demo():

    import sys
    from pathlib import Path

    parent_dir = Path(__file__).resolve().parents[2]
    print(parent_dir)
    sys.path.append(parent_dir)



"""
LINKS
are atomic LLM interactions. They transform input data into a prompt, send it to an LLM model, parse the output, and return the result.
"""

import logging
from abc import ABC, abstractmethod
from typing import Callable, Iterable

from fence.models.base import LLM
from fence.models.claude3 import Claude3Base
from fence.models.openai import GPTBase
from fence.parsers import Parser
from fence.templates import MessagesTemplate, StringTemplate
from fence.utils.base import time_it

logger = logging.getLogger(__name__)

PROTECTED_LINK_KWARGS = ["model"]

################
# Base classes #
################


class BaseLink(ABC):
    """
    BaseLink class is a simple input-output building block. It is callable, and will call the run method. The run method
    takes a dictionary of input variables for the template, and returns a dictionary with the output under the specified
    key.
    """

    def __init__(self, output_key: str = "state", name: str = None):
        """
        Initialize the BaseLink object.

        :param input_keys: A list of input keys.
        :param output_key: The output key.
        """
        self.model = None
        self.input_keys = None
        self.output_key = output_key
        self.name = name

    def __call__(self, **kwargs):
        """
        Alias for the run method to allow calling an instance as a function.
        :param kwargs: Keyword arguments to be passed to the run method.
        :return: Output of the run method.
        """
        return self.run(**kwargs)

    @abstractmethod
    def run(self, input_dict: dict = None, **kwargs):
        """
        Run the link.
        :param input_dict: Values for the input variables.
        :param kwargs: Additional keyword arguments for the LLM model.
        :return:
        """
        raise NotImplementedError(
            "Base class <BaseLink> `run` method must be implemented."
        )

    def _validate_input(self, input_dict: dict):
        """
        Validate the input dictionary. Input dictionary should contain all the input keys for the Link.
        :param input_dict: Input dictionary.
        :return: True if the input dictionary is valid, False otherwise.
        """
        # Find both missing and superfluous variables
        missing_variables = [
            variable for variable in self.input_keys if variable not in input_dict
        ]
        superfluous_variables = [
            variable for variable in input_dict if variable not in self.input_keys
        ]
        if missing_variables:
            raise ValueError(f"Missing variables: {missing_variables}")
        if superfluous_variables:
            logger.debug(f"Superfluous variables: {superfluous_variables}")

    def __str__(self):
        return f"Link: {f'{self.name} ' if self.name else ''}<{self.input_keys}> -> <{self.output_key}>"

    def __repr__(self):
        return str(self)


###################
# Implementations #
###################


class TransformationLink(BaseLink):
    def __init__(
        self,
        input_keys: Iterable[str],
        output_key: str = "state",
        name: str = None,
        function: Callable = None,
    ):
        """
        Initialize the TransformationLink object.

        :param input_keys: A list of input keys.
        :param output_key: The output key.
        :param function: The transformation function.
        """
        super().__init__(output_key=output_key, name=name)
        self.input_keys = input_keys
        self.function = function

    @time_it
    def run(self, input_dict: dict = None, **kwargs):
        """
        Run the link.
        :param input_dict: Values for the input variables.
        :param kwargs: Additional keyword arguments for the LLM model.
        :return:
        """
        logger.info(
            f"Executing {f'<{self.name}> ' if self.name else ''}Transformation Link"
        )

        # Validate the input dictionary
        self._validate_input(input_dict)

        # Call the transformation function
        response = self.function(input_dict)
        logger.debug(f"Response: {response}")

        # Build the response dictionary #
        #################################

        # We always return the output under the key 'output', but if the output key is different, we also return it
        # under that key. This allows us to track the output of a specific Link in a Chain, while still having a
        # consistent key for the output of the Link that allows subsequent Links to find the latest output.
        response_dict = {"state": response}
        if not self.output_key == "state":
            response_dict[self.output_key] = response
        return response_dict


class Link(BaseLink):
    """
    Link class that takes an LLM model, and calls it using the prompt template. Base blocks for building more complex
    chains. The Link class is callable, and will call the run method. The run method takes a dictionary of input
    variables for the template, and returns a dictionary with the output of the LLM model under the specified key.
    """

    def __init__(
        self,
        template: StringTemplate | MessagesTemplate,
        output_key: str = "state",
        model: LLM = None,
        name: str = None,
        parser: Parser = None,
    ):
        """
        Initialize the Link object.

        :param name: Name of the Link.
        :param model: An LLM model object.
        :param template: A PromptTemplate object.
        """
        self.template = template
        super().__init__(output_key=output_key, name=name)
        self.model = model
        self.template = template

        # If the template is a MessagesTemplate, we only accept Claude3 type models
        if isinstance(template, MessagesTemplate):
            if not isinstance(model, (Claude3Base, GPTBase)):
                logger.warning(
                    f"MessagesTemplate can only be used with Claude3 or GPT models. Got {model.model_id if model else None}."
                )

                # Reformat the template to a StringTemplate
                self.template = self.template.to_string_template()

        # Get the input keys from the template
        self.input_keys = template.input_variables
        self.parser = parser

    def __call__(self, **kwargs):
        """
        Alias for the run method to allow calling an instance as a function.
        :param kwargs: Keyword arguments to be passed to the run method.
        :return: Output of the run method.
        """
        return self.run(**kwargs)

    @time_it(threshold=30, only_warn=True)
    def run(self, input_dict: dict = None, **kwargs):
        """
        Run the link.
        :param input_dict: Variables for the template.
        :param kwargs: Additional keyword arguments for the LLM model.
        :return:
        """

        logger.info(f"Executing {f'<{self.name}>' if self.name else 'unnamed'} Link")

        # Update the input dictionary with the keyword arguments
        if input_dict is None:
            input_dict = {}

        input_dict.update(kwargs)

        # Check if an LLM model was provided
        if self.model is None and kwargs.get("model") is None:
            raise ValueError(
                "An model must be provided, either as an argument or in the Link object."
            )

        # Validate the input dictionary
        self._validate_input(input_dict)

        # Render the template
        prompt = self.template.render(input_dict=input_dict)
        logger.debug(f"Prompt: {prompt}")

        # Determine if the LLM model is provided as a keyword argument,
        # otherwise use the LLM model of the Link
        model = input_dict.pop("model", self.model)

        # Call the LLM model
        response = model.invoke(prompt=prompt)
        logger.debug(f"Raw response: {response}")

        # Parse the response
        if self.parser is not None:
            response = self.parser.parse(response)
            logger.debug(f"Parsed response: {response}")

        # Build the response dictionary #
        #################################

        # We always return the output under the key 'output', but if the output key is different, we also return it
        # under that key. This allows us to track the output of a specific Link in a Chain, while still having a
        # consistent key for the output of the Link that allows subsequent Links to find the latest output.

        # If the response is a dictionary, we assume it contains the primary output under the key 'state'
        if isinstance(response, dict):
            if "state" not in response.keys():
                logger.debug(
                    "Response is a dictionary, but does not contain the key "
                    "'state'. Using the full response as the state."
                )
                response_dict = {"state": response}
            else:
                response_dict = {"state": response["state"]}
        elif isinstance(response, (str, bool)):
            response_dict = {"state": response}
        else:
            raise TypeError(
                f"Response must be a dictionary, string or bool. Got {type(response)}."
            )

        # If the output key is not 'state', we also return the *full* output under that key
        if not self.output_key == "state":
            response_dict[self.output_key] = response

        logger.debug(f"Current state: {response_dict['state']}")

        return response_dict



import json

from fence.tools.base import BaseTool


class EnvTool(BaseTool):
    """Tool to access environment variables"""

    def execute_tool(self, **kwargs) -> str:
        """Print the environment"""
        environment = kwargs.get("environment", {})
        env_vars = ", ".join([f"{k}: {v}" for k, v in environment.items()])
        return json.dumps(
            f"The environment currently holds these variables:\n) {env_vars}"
        )



"""
This module contains the MessagesTemplate class, which is used to render a list of messages with the provided variables.
The module was created to accommodate the Claude3 model API format.
"""

from fence.templates.base import BaseTemplate
from fence.templates.models import Message, Messages, TextContent
from fence.templates.string import StringTemplate


class MessagesTemplate(BaseTemplate):

    def __init__(self, source: str | Messages, **kwargs):
        """
        Initialize the MessagesTemplate object.

        :param str | Messages source: The template string, or a list of Message objects.
        :param list input_variables: List of input variables expected by the template.
        :param kwargs: Additional keyword arguments.
            - separator (str, optional): Separator to use when merging string-based templates. Default is a space.
        """

        super().__init__(source=source)

        # Find placeholders in the template
        self.input_variables = self._find_placeholders()

    def render(self, input_dict: dict = None, **kwargs):
        """
        Render a list of messages with the provided variables.

        :param Messages messages: List of Message objects.
        :param dict input_dict: Dictionary of input variables.
        :param kwargs: Keyword arguments containing values for the template variables.
        :return: Rendered list of messages.
        :rtype: Messages

        :raises ValueError: If any required variable is missing.
        """

        # Make input_dict
        if input_dict is None:
            input_dict = {}
        input_dict.update(kwargs)

        # Find both missing and superfluous variables
        self._validate_input(input_dict=input_dict)

        # Messages are the source
        messages = self.source

        # First get the system message, if any, and render it
        system = None
        if messages.system:
            system = self._render_string(messages.system, input_dict=input_dict)

        # Then go through each user/assistant message
        rendered_messages = []

        for message in messages.messages:

            # If content is a string, render it
            if isinstance(message.content, str):
                content = self._render_string(message.content, input_dict=input_dict)

            # If content is a list of content objects...
            elif isinstance(message.content, list):

                # ...go through each content object and render text content
                content = []
                for content_object in message.content:
                    if content_object.type == "text":
                        content.append(
                            TextContent(
                                text=self._render_string(
                                    content_object.text, input_dict=input_dict
                                )
                            )
                        )
                    elif content_object.type == "image":
                        content.append(content_object)

            # If content is neither a string nor a list, raise an error
            else:
                raise TypeError(
                    "Content must be a string or a list of content objects."
                )

            rendered_messages.append(Message(role=message.role, content=content))

        return Messages(messages=rendered_messages, system=system)

    # Helper functions
    def _find_placeholders(self) -> list[str]:
        """
        Find placeholders in a list of messages. Placeholders are defined as text enclosed in double curly braces.
        They are used to denote variables that need to be replaced in the messages.

        :param Messages messages: The list of messages.
        :return: List of placeholders.
        :rtype: list[str]
        """

        # Initialize placeholders
        placeholders = set()

        # ...otherwise, the template is a Messages object
        messages = self.source

        # Check the system message
        if messages.system:
            placeholders.update(self._find_placeholders_in_text(text=messages.system))

        # Then go through each user/assistant message
        for message in messages.messages:

            # Content can be a string, or a list of content objects. If it's a list...
            if isinstance(message.content, list):

                # ...go through each content object and find placeholders in text content
                for content in message.content:
                    if content.type == "text":
                        placeholders.update(
                            self._find_placeholders_in_text(text=content.text)
                        )

            # If content is a string, find placeholders in the string
            elif isinstance(message.content, str):
                placeholders.update(
                    self._find_placeholders_in_text(text=message.content)
                )

            # If content is neither a string nor a list, raise an error
            else:
                raise TypeError(
                    "Content must be a string or a list of content objects."
                )

        return list(placeholders)

    def to_string_template(self) -> StringTemplate:
        """
        Convert the MessagesTemplate object to a StringTemplate object.

        :return: StringTemplate object.
        :rtype: StringTemplate
        """

        # If the source is a string, return a StringTemplate object
        if isinstance(self.source, str):
            return StringTemplate(source=self.source)

        # If the source is a Messages object, convert it to a string
        messages = self.source

        # Initialize the string
        new_source = ""

        # Add the system message
        if messages.system:
            new_source += messages.system + "\n"

        # Add the user/assistant messages
        for message in messages.messages:
            if isinstance(message.content, str):
                new_source += message.content + "\n"
            elif isinstance(message.content, list):
                for content in message.content:
                    if content.type == "text":
                        new_source += content.text + "\n"

        return StringTemplate(source=new_source)

    def __str__(self):
        return f"MessagesTemplate: {self.source}"

    def __repr__(self):
        return f"MessagesTemplate(template={self.source}, input_variables={self.input_variables})"

    def __add__(self, other):
        """
        Combine two MessagesTemplate instances.
        :param other: other MessagesTemplate instance
        :return: Combined MessagesTemplate instance
        """

        if not isinstance(other, MessagesTemplate):
            raise TypeError("Can only combine MessagesTemplate instances.")

        # Combine the messages
        system = self.source.system + " " + other.source.system
        messages = self.source.messages + other.source.messages

        return MessagesTemplate(Messages(system=system, messages=messages))

    def __eq__(self, other):
        """
        Check if two MessagesTemplate instances are equal.
        :param other: other MessagesTemplate instance
        :return: True if the two instances are equal, False otherwise
        """

        if not isinstance(other, MessagesTemplate):
            return False

        return (
            self.source == other.source
            and self.input_variables == other.input_variables
        )



from .base import LLM  # noqa: F401
from .claude import ClaudeBase, ClaudeInstant, ClaudeV2  # noqa: F401
from .claude3 import Claude35Sonnet, ClaudeHaiku, ClaudeSonnet  # noqa: F401
from .ollama import Llama3_1, Ollama, OllamaBase  # noqa: F401
from .openai import GPT4o, GPTBase  # noqa: F401



"""
Agent class to orchestrate a flow that potentially calls tools or other, specialized agents.
"""

import logging

from fence import LLM, Link, TOMLParser
from fence.agents.base import BaseAgent
from fence.links import logger as link_logger
from fence.memory.base import BaseMemory, FleetingMemory
from fence.memory.dynamodb import DynamoDBMemory
from fence.models.openai import GPT4omini
from fence.prompts.agents import REACT_MULTI_AGENT_TOOL_PROMPT
from fence.tools.base import BaseTool

logger = logging.getLogger(__name__)

# Suppress the link logger
link_logger.setLevel("DEBUG")


class SuperAgent(BaseAgent):
    """An LLM-based agent capable of delegating tasks to other agents or tools."""

    def __init__(
        self,
        identifier: str | None = None,
        model: LLM | None = None,
        description: str | None = None,
        role: str | None = None,
        delegates: list[BaseAgent] | None = None,
        tools: list[BaseTool] | None = None,
        memory: BaseMemory | None = None,
        environment: dict | None = None,
    ):
        """
        Initialize the Agent object.

        :param identifier: An identifier for the agent. If none is provided, the class name will be used.
        :param model: An LLM model object.
        :param description: A description of the agent. Used to represent the agent to other agents.
        :param role: The role of the agent. Used to represent the agent to itself, detailing its purpose.
        :param delegates: A list of delegate agents.
        :param tools: A list of Tool objects.
        :param memory: A memory class. Defaults to FleetingMemory.
        :param environment: A dictionary of environment variables to pass to delegates and tools.
        """
        super().__init__(
            identifier=identifier,
            model=model,
            description=description,
            environment=environment or {},
        )

        # Set role
        self.role = (
            role
            or "You are a general purpose agent, capable of delegating tasks to other agents or tools."
        )

        # Delegates and tools initialization
        self.delegates = (
            {delegate.identifier: delegate for delegate in delegates}
            if delegates
            else {}
        )
        # Add environment variables to delegates
        for delegate in self.delegates.values():
            delegate.environment.update(self.environment)

        # Initialize tools with environment variables
        if tools:
            for tool in tools:
                tool.environment.update(self.environment)
            self.tools = {tool.__class__.__name__: tool for tool in tools}
        else:
            self.tools = {}

        # Logging agent creation details
        logger.info(
            f"Initialized agent with model <{model.model_name}>, delegates: {list(self.delegates.keys())}, tools: {list(self.tools.keys())}"
        )

        # Memory setup
        self.memory = memory or FleetingMemory()

        # Prepare formatted delegates and tools
        self.formatted_delegates = self._format_entities(list(self.delegates.values()))
        self.formatted_tools = self._format_entities(list(self.tools.values()))

        # Initialize the memory buffer
        self.wipe_memory()

    def run(self, prompt: str) -> str:
        """
        Main execution loop to process the input prompt.

        :param prompt: The input prompt to process.
        :return: The final answer from the agent.
        """

        # Add initial prompt to the memory buffer
        self.memory.add_message(role="user", content=prompt)

        max_iterations, iteration_count = 30, 0
        response, answer = None, None

        while iteration_count < max_iterations:

            link = Link(
                name=f"{self.identifier or 'agent'}_step_{iteration_count}",
                model=self.model,
                template=self.memory.to_messages_template(),
            )
            response = link.run()["state"]
            logger.info(f"Model response: {response}")

            self.memory.add_message(role="assistant", content=response)

            if "[ANSWER]" in response:
                answer = self._extract_answer(response)
                break

            if "[DELEGATE]" in response:
                delegate_response = self._handle_delegate_action(response)
                if delegate_response:
                    self.memory.add_message(
                        role="user", content=f"[OBSERVATION] {delegate_response}"
                    )

            if "[ACTION]" in response:
                tool_response = self._handle_tool_action(response)
                if tool_response:
                    self.memory.add_message(
                        role="user", content=f"[OBSERVATION] {tool_response}"
                    )

            iteration_count += 1

        self.wipe_memory()
        return answer or "No answer found"

    def wipe_memory(self):
        """Clear or reset the agent's memory context."""

        # Check if there are any messages in the memory
        self.memory.messages = self.memory.get_messages()

        # Check if there is a system message in the memory
        self.memory.system = self.memory.get_system_message()

        # If no system message is present, add a new one
        if not self.memory.system:
            self.memory.set_system_message(
                REACT_MULTI_AGENT_TOOL_PROMPT.format(
                    role=self.role,
                    delegates=self.formatted_delegates,
                    tools=self.formatted_tools,
                )
            )

    def _format_entities(self, entities: list) -> str:
        """Format delegates or tools into TOML representation."""
        return (
            "".join(entity.format_toml() for entity in entities)
            if entities
            else "None available"
        )

    def _extract_answer(self, response: str) -> str:
        """Extract the final answer from the response."""
        return response.split("[ANSWER]")[-1].strip()

    def _handle_delegate_action(self, response: str) -> str:
        """Handle actions involving delegate agents."""

        delegate_block = response.split("[DELEGATE]")[-1].strip()
        logger.debug(f"Processing delegate: {delegate_block}")

        try:
            delegate_data = TOMLParser().parse(input_string=delegate_block)
            delegate_name = delegate_data["delegate_name"]
            delegate_input = delegate_data["delegate_input"]
            logger.debug(
                f"Executing delegate '{delegate_name}' with input: {delegate_input}"
            )

            if delegate_name in self.delegates:
                delegate = self.delegates[delegate_name]
                return delegate.run(prompt=delegate_input)
            else:
                logger.error(f"Delegate {delegate_name} not found.")
                return "Delegate not found"
        except Exception as e:
            logger.error(f"Error processing delegate: {e}")
            return "Delegate execution failed"

    def _handle_tool_action(self, response: str) -> str:
        """Handle actions involving tools."""

        action_block = response.split("[ACTION]")[-1].strip()
        logger.debug(f"Processing tool action: {action_block}")

        try:
            action_data = TOMLParser().parse(input_string=action_block)
            tool_name = action_data.get("tool_name")
            tool_params = action_data.get("tool_params", {})
            logger.info(f"Executing tool '{tool_name}' with params: {tool_params}")

            if tool_name in self.tools:
                tool = self.tools[tool_name]
                tool_response = tool.run(**tool_params, environment=self.environment)
                logger.info(f"Tool <{tool_name}> response: {tool_response}")
                return tool_response
            else:
                logger.error(f"Tool {tool_name} not found.")
                return "Tool not found"
        except Exception as e:
            logger.error(f"Error processing tool action: {e}")
            return "Tool execution failed"


if __name__ == "__main__":

    # # Create the delegate agent
    # delegate = ToolAgent(
    #     model=GPT4omini(source="test"),
    #     description="An swiss army knife agent, with math and basic string manipulation tools",
    #     tools=[CalculatorTool(), PrimeTool()],
    # )
    #
    # # Create an intermediary referee agent
    # intermediary = SuperAgent(
    #     model=GPT4omini(source="test"),
    #     delegates=[delegate],
    #     tools=[SecretStringTool()],
    #     description="An intermediary agent that can has access to a delegate agent with many tools, and that has access to a secret string tool",
    # )
    #
    # # Create a chat agent
    # chat_agent = ChatAgent(
    #     identifier="Rephraser",
    #     model=GPT4omini(source="chat"),
    #     description="A chat agent that rephrases things",
    #     profile="You are a chat agent with a snarky attitude. You are here to provide witty responses to user queries. Give me an assignment, and I will complete it with a snarky twist.",
    # )
    #
    # # Create the referee agent
    # master = SuperAgent(
    #     model=GPT4omini(source="test"),
    #     delegates=[intermediary, chat_agent],
    #     tools=[TextInverterTool()],
    # )
    #
    # # Trigger the agent
    # master.run(
    #     prompt="What is the secret string? Then, invert it. Finally, ask for advice on how to best phrase this. It is important the inverted string is in the final response."
    # )

    # # Define the tools available to the agent
    # tools = [CalculatorTool(), PrimeTool(), TextInverterTool(), EnvTool()]
    #
    # # Create an agent with a model and tools
    # agent = SuperAgent(
    #     model=GPT4omini(source="agent"),
    #     tools=tools,
    #     environment={"some_env_var": "some_value"},
    # )
    #
    # for q in [
    #     # "How much is 9 + 10?",
    #     # "Is 1172233 a prime number?",
    #     # "What is the square root of 16?",
    #     # Math question we don't have a tool for
    #     # "Find the first 2 prime numbers beyond 10000",
    #     # "Find the sum of the first 2 prime numbers beyond 1005, take the number as a string and reverse it",
    #     "Tell me what the value of the environment variable 'some_env_var' is",
    # ]:
    #     logger.critical(f"Running agent with prompt: {q}")
    #     response = agent.run(q)
    #     logger.critical(f"Response: {response}")

    class AccountNameRetrieverTool(BaseTool):
        """
        Tool to retrieve the account holder name from a database.
        """

        def execute_tool(self, environment):
            account_id = self.environment.get("current_account_id", "unknown")
            logger.info(f"Retrieving account holder name for account_id: {account_id}")
            if account_id == "foo":
                return "Bert"
            if account_id == "bar":
                return "Ernie"
            return "Unknown"

    # Create the memory object
    memory = DynamoDBMemory(
        table_name="fence_test",
        primary_key_name="session",
        primary_key_value="02d2b1b0-cf84-401e-b9d9-16f24c359cc8",
    )

    # Create the agents
    child_agent = SuperAgent(
        identifier="child_accountant",
        description="An agent that can retrieve the account holder name",
        model=GPT4omini(source="agent"),
        tools=[AccountNameRetrieverTool()],
    )
    parent_agent = SuperAgent(
        identifier="parent_accountant",
        model=GPT4omini(source="agent"),
        delegates=[child_agent],
        environment={"current_account_id": "bar"},
        memory=memory,
    )
    result = parent_agent.run("what is the current account holders name?")



"""
Memory classes.
"""

import logging
import uuid
from datetime import datetime, timezone

import boto3

from fence import Message, Messages
from fence.memory.base import BaseMemory

logger = logging.getLogger(__name__)


class DynamoDBMemory(BaseMemory):
    """
    A dynamoDB-based memory context for the agent.
    """

    def __init__(
        self,
        table_name: str,
        primary_key_name: str = "PK",
        primary_key_value: str | None = None,
        primary_key_value_prefix: str = "",
        sort_key_name: str = "SK",
        region_name: str = "eu-central-1",
        source: str = None,
    ):
        """
        Initialize the DynamoDBMemory object.
        :param str table_name: The name of the DynamoDB table.
        :param str primary_key_name: The name of the partition key.
        :param str primary_key_value: The partition key.
        :param str primary_key_value_prefix: The prefix for the partition key. Useful for tenant isolation.
        :param str sort_key_name: The name of the sort key.
        :param str region_name: The AWS region name.
        :param str source: A custom identifier for what created the memory object.
        """

        self.table_name = table_name
        self.source = source

        # Connect to the DynamoDB service
        try:
            self.db = boto3.resource("dynamodb", region_name=region_name)
            self.table = self.db.Table(self.table_name)
        except Exception as e:
            logger.error(
                f"Failed to connect to DynamoDB table {table_name} - {e}", exc_info=True
            )
            raise ValueError(f"Failed to connect to DynamoDB table {table_name}")

        # Set PK and SK
        self.primary_key_name = primary_key_name
        self.sort_key_name = sort_key_name

        # Primary key counts as session ID
        self.primary_key_value = (
            primary_key_value
            if primary_key_value
            else f"{primary_key_value_prefix}{uuid.uuid4()}"
        )

        # Check if the table exists
        if self.table.table_status != "ACTIVE":
            raise ValueError(f"Table {table_name} is not active")

    def add_message(self, role: str, content: str, meta: dict | None = None):
        """
        Add a message to the memory buffer.
        :param role: The role of the message.
        :param content: The content of the message.
        :param meta: The meta information of the message.
        """

        # Store the message in DynamoDB
        dynamo_response = self.table.put_item(
            Item=self._format_message_for_dynamo_db(
                role=role, content=content, meta=meta
            )
        )
        status_code = dynamo_response["ResponseMetadata"]["HTTPStatusCode"]
        if status_code != 200:
            logger.error(f"Failed to store message in DynamoDB: {dynamo_response}")
            raise ValueError(f"Failed to store message in DynamoDB: {dynamo_response}")

    def add_user_message(self, content: str):
        self.add_message(role="user", content=content)

    def add_assistant_message(self, content: str):
        self.add_message(role="assistant", content=content)

    def set_system_message(self, content: str):
        self.add_message(role="system", content=content)

    def add_messages(self, messages: list[Message]):
        """
        Add a list of messages to the memory buffer.
        :param messages: The list of messages.
        """
        for message in messages:
            self.add_message(role=message.role, content=message.content)

    def _get_messages(self):
        """
        Get ALL messages from the dynamoDB.
        """

        logger.info(f"Getting history for session {self.primary_key_name}")
        response = self.table.query(
            KeyConditionExpression=boto3.dynamodb.conditions.Key(
                f"{self.primary_key_name}"
            ).eq(self.primary_key_value)
        )
        items = response.get("Items", [])
        messages = Messages(
            messages=[
                Message(
                    role=item["Role"],
                    content=item["Message"],
                )
                for item in items
                if item["Role"] in ["user", "assistant"]
            ]
        )
        messages.system = next(
            (item["Message"] for item in items if item["Role"] == "system"),
            None,
        )

        # Log the messages
        logger.info(
            "Messages:\n"
            + "\n".join(
                [
                    f"<{message.role.upper()}>:\t{message.content}"
                    for message in messages.messages
                ]
            )
        )

        return messages

    def get_messages(self):
        """
        Get the messages from the memory buffer.
        """
        messages = self._get_messages()
        return messages.messages

    def get_system_message(self):
        """
        Get the system message from the memory buffer.
        """
        messages = self._get_messages()
        return messages.system

    ###########
    # Helpers #
    ###########

    def _format_message_for_dynamo_db(
        self,
        role: str,
        content: str,
        meta: dict | None = None,
    ) -> dict:
        """
        Format a message for storage in DynamoDB.
        :param role: The role of the message.
        :param content: The content of the message.
        :param meta: The meta information of the message.
        :return: The formatted message.
        """
        return {
            f"{self.primary_key_name}": self.primary_key_value,
            f"{self.sort_key_name}": datetime.now(timezone.utc).isoformat(),
            "Message": content,
            "Role": role,
            "Meta": meta,
            "Source": self.source,
        }


if __name__ == "__main__":

    # Create a DynamoDBMemory object
    memory = DynamoDBMemory(
        table_name="fence_test",
        primary_key_name="session",
        primary_key_value="08f1201a-d9cd-4ef7-9b63-c26d2c824576",
        # primary_key_value="test_a",
    )

    # # Add a system message
    # memory.set_system_message("This is a system message")
    #
    # # Add a user message
    # memory.add_user_message("This is a user message")
    #
    # # Add an assistant message
    # memory.add_assistant_message("This is an assistant message")

    # Print the messages
    messages = memory.get_messages()
    system = memory.get_system_message()
    logger.info(f"System message <{type(system)}>: {system}")
    logger.info(f"Messages <{type(messages)}>: {messages}")



"""
Helper tools to perform basic LLM-related tasks

- Relevancy check: Check the relevancy of a text to a given topic
- TODO: Consistency/contradiction check: Check the contradiction of two texts
"""

import logging
from math import ceil

from fence import (
    BoolParser,
    Claude3Base,
    ClaudeHaiku,
    Link,
    Message,
    Messages,
    MessagesTemplate,
)
from fence.parsers import TripleBacktickParser
from fence.utils.optim import retry

logger = logging.getLogger(__name__)

DEFAULT_REFERENCE_FLAGS = ["according", "text", "passage"]

##########
# String #
##########


def get_first_n_words(text: str, n: int) -> str:
    """
    Get the first n words from a text
    """
    return " ".join(text.split()[:n])


def get_word_count(text: str) -> int:
    """
    Get the word count of a text
    """
    return len(text.split())


############
# Chunking #
############


class TextChunker:
    def __init__(
        self,
        chunk_size: int,
        overlap: float | None = 0.01,
        text: str = None,
    ):
        self.text = text
        self.chunk_size = chunk_size
        self.overlap = overlap

    @staticmethod
    def _get_approximate_chunk_size(text: str, chunk_size: int) -> int:
        """
        Get approximate chunk size for a given text and chunk size
        :param text: text string
        :param chunk_size: desired chunk size target
        :return: recalculated chunk size
        """
        approx_chunk_count = ceil(len(text) / chunk_size)
        approx_chunk_size = ceil(len(text) / approx_chunk_count)
        return approx_chunk_size

    def split_text(self, text: str = None) -> list[str]:
        text = text if text is not None else self.text

        core_chunk_size = self._get_approximate_chunk_size(
            text=text, chunk_size=int(self.chunk_size / (1 + self.overlap))
        )

        chunk_size_with_overlap = (
            (core_chunk_size + int(core_chunk_size * self.overlap))
            if self.overlap
            else core_chunk_size
        )

        logger.debug(
            f"Chunking {len(text)} characters into {self.chunk_size=}-sized chunks with : {core_chunk_size=}, {chunk_size_with_overlap=}"
        )

        remaining_text = text
        chunks = []

        while len(remaining_text) > core_chunk_size:
            chunk = (
                remaining_text[:chunk_size_with_overlap]
                if len(remaining_text) > chunk_size_with_overlap
                else remaining_text
            )
            chunks.append(chunk)
            remaining_text = remaining_text[core_chunk_size:]

        if remaining_text:
            chunks.append(remaining_text)

        logger.debug(
            f"Chunking complete. {len(chunks)} chunks created. Smallest chunk size: {min([len(chunk) for chunk in chunks])}"
        )

        return chunks


#######
# LLM #
#######


class LLMHelper:
    """
    Helper class to perform basic LLM tasks
    """

    def __init__(self, model: Claude3Base = None):
        self.model = model if model else ClaudeHaiku(source="llm_helper")

    @retry(max_retries=3)
    def check_relevancy(self, text: str, topic: str) -> bool:
        """
        Check the relevancy of a text to a given topic
        :param text: The text to check
        :param topic: The topic to check against
        """

        # Define template
        relevancy_template = MessagesTemplate(
            source=Messages(
                system="""You are a helpful search assistant. Your job is to judge
                whether a given text is relevant to a given topic. You are given a text
                and a topic. You should return True if the text is relevant to the
                topic, and False otherwise.""",
                messages=[
                    Message(
                        role="user",
                        content="""You are given the following text: {{text}}
                    The topic is: {{topic}}
                    Is the text relevant to the topic? Only reply with True or False, and nothing else.
                    """,
                    ),
                    Message(
                        role="assistant",
                        content="Evaluation boolean (True/False):",
                    ),
                ],
            )
        )
        logger.debug(
            f"Checking relevancy of text <{get_first_n_words(text=text, n=10)}...> to topic <{topic}>"
        )
        relevancy_link = Link(
            template=relevancy_template,
            model=self.model,
            name="relevancy_checker",
            output_key="is_relevant",
            parser=BoolParser(),
        )
        response = relevancy_link.run({"text": text, "topic": topic})

        # Return check
        relevancy_check = response["is_relevant"]
        logger.debug(f"Relevancy check response: {relevancy_check}")

        return relevancy_check

    @retry(max_retries=3)
    def remove_text_references(
        self, text: str, reference_flags: list | None = None
    ) -> str:
        """
        Check if the given text contains any of the reference flags, and clean them up
        :param text: The text to check
        :param reference_flags: The list of reference flags to check against
        """

        # Set default reference flags
        reference_flags = (
            reference_flags if reference_flags else DEFAULT_REFERENCE_FLAGS
        )

        # If the text does not contain any of the reference flags, return the text as is
        if not any(flag in text.lower() for flag in reference_flags):
            return text

        # Define template
        rewrite_template = MessagesTemplate(
            source=Messages(
                system="""You are a helpful editor. Your job is to make sure that text you are given
                 does not contain any references to a text or passage, since that is something that
                 the reader cannot see. The answer must stand on its own. If it already does, you should
                    reply with the same text. Otherwise, you should remove any references to the text or passage.
                    Reply with the edited or original text, delimited by triple backticks. Make sure to use proper spelling and grammar.

                 EXAMPLE:

                 Input: According to the text, the sky is blue because of Rayleigh scattering, which is the scattering of light by air molecules.
                 Output: The sky is blue because of Rayleigh scattering, which is the scattering of light by air molecules.

                 Input: The key reason given in the text for not using unit tests is that they are time-consuming.
                 Output: The key reason for not using unit tests is that they are time-consuming.
                 """,
                messages=[
                    Message(
                        role="user",
                        content="""Input text:```{text}```
                                """,
                    ),
                    Message(
                        role="assistant",
                        content="Revised text: ```",
                    ),
                ],
            )
        )
        logger.debug(
            f"Checking text <{get_first_n_words(text=text, n=10)}...> for references <{reference_flags}>"
        )
        rewrite_link = Link(
            template=rewrite_template,
            model=self.model,
            name="text_cleaner",
            output_key="cleaned_text",
            parser=TripleBacktickParser(prefill="```"),
        )
        response = rewrite_link.run({"text": text})

        # Return cleaned text
        cleaned_text = response["cleaned_text"]
        logger.debug(f"Cleaned text response: {cleaned_text}")

        return cleaned_text



"""
Tools for agents
"""

import logging
from abc import ABC, abstractmethod

logger = logging.getLogger(__name__)


class BaseTool(ABC):

    def __init__(self, description: str = None, environment: dict = None):
        """
        Initialize the BaseTool object.

        :param description: A description of the tool (if not provided, the docstring will be used)
        :param environment: A dictionary of environment variables to pass to the tool
        """
        self.description = description
        # self.environment = environment or {}
        self.environment = {}

    def run(self, environment: dict = None, **kwargs):
        """
        Base method that is called when the tool is used.
        This method will always accept 'environment' as a parameter.
        """

        # Initialize environment if not provided
        self.environment = environment or self.environment

        # Preprocess environment if needed
        # self.preprocess_environment(environment)

        # Pass the environment to the execute_tool method
        kwargs["environment"] = environment

        # Call the subclass-specific implementation
        return self.execute_tool(**kwargs)

    @abstractmethod
    def execute_tool(self, environment: dict = None, **kwargs):
        """
        The method that should be implemented in the derived class.
        This will handle the actual tool-specific logic.

        :param environment: A dictionary of environment variables to pass to the tool
        :param kwargs: Additional keyword arguments for the tool
        """
        raise NotImplementedError

    # def preprocess_environment(self, environment: dict):
    #     """
    #     Optional method to preprocess the environment, can be overridden by subclasses.
    #     """
    #     pass

    def format_toml(self):
        """
        Returns a TOML-formatted key-value pair of the tool name,
        the description (docstring) of the tool, and the arguments
        of the `run` method.
        """
        # Get the arguments of the run method
        run_args = self.execute_tool.__annotations__
        run_args.pop("return", None)
        run_args.pop("environment", None)

        # Preformat the arguments
        argument_string = ""
        if run_args:
            for arg_name, arg_type in run_args.items():
                argument_string += (
                    f"[[tools.tool_params]]\n"
                    f'name = "{arg_name}"\n'
                    f'type = "{arg_type.__name__}"\n'
                )
        else:
            argument_string = "# No arguments"

        # Get the description of the tool, if not provided, use the docstring
        tool_description = self.description or self.__doc__

        if not tool_description:
            logger.warning(
                f"Tool {self.__class__.__name__} has no description or docstring."
            )

        # Format the TOML representation of the tool for the agent
        toml_string = f"""[[tools]]
tool_name = "{self.__class__.__name__}"
tool_description = "{tool_description}"
{argument_string}"""

        return toml_string



import logging

from fence import Claude35Sonnet
from fence.agents.chat import ChatAgent
from fence.models.openai import GPT4omini

logger = logging.getLogger(__name__)


class ChatAgency:
    """
    An Agency object, capable of managing multiple Agents.

    It starts off with an initial prompt, after which each agent runs in turn, taking the output of the previous agent as input.
    """

    def __init__(self, agents: list[ChatAgent] = None):
        """
        Initialize the Agency object.

        :param agents: A list of Agent objects.
        """
        self.agents = agents or []
        logger.info(
            f"Creating an agency with agents: {[agent.name for agent in self.agents]}"
        )

    def run(self, prompt: str, max_rounds: int = 3) -> str:
        """
        Run the agency with the given prompt.

        :param prompt: The initial prompt to feed to the first agent.
        :param max_rounds: The maximum number of rounds of conversation.
        :return: The final response from the last agent.
        """
        if not self.agents:
            return "No agents available to process the prompt."

        # Log the initial prompt
        logger.info(f"Opening prompt: {prompt}")

        # Initialize the context
        context = prompt

        # Start the conversation
        round_count = 0
        while round_count < max_rounds:

            # Loop through the agents, where every agent has all the previous agents' responses as context
            for agent in self.agents:

                # Run the agent
                context = agent.run(context)

            # Increment the round count
            round_count += 1

        return context


if __name__ == "__main__":

    # Create 3 ChatAgents
    agent1 = ChatAgent(
        model=GPT4omini(
            source="chat",
        ),
        name="GPT-4o-mini",
        profile="You are GPT-4o-mini. You pretend to be a hipster, but secretly you are as corporate as they come. You are edgy and slightly confrontational.",
    )
    agent2 = ChatAgent(
        model=Claude35Sonnet(
            source="chat",
        ),
        name="Claude35Sonnet",
        profile="You are Claude35Sonnet. You are value ethics and privacy. You are courteous, but you don't trust OpenAI. You are quick-witted and have a dry sense of humor.",
    )

    # agent3 = ChatAgent(
    #     model=Ollama(
    #         model_id='llama3.1',
    #         source="chat",
    #     ),
    #     profile="You are Llama3. You are adamant about the importance of open source software. Corporate greed is your nemesis."
    # )

    # Create the ChatAgency
    agency = ChatAgency(agents=[agent1, agent2])

    # Run the agency
    response = agency.run("Today's topic is manga!", max_rounds=10)



"""
Memory classes.
"""

import logging
from abc import ABC, abstractmethod

from fence import Message, Messages, MessagesTemplate

logger = logging.getLogger(__name__)


class BaseMemory(ABC):
    """
    Base memory class for agents.
    """

    ###########
    # Setters #
    ###########

    @abstractmethod
    def add_message(self, role: str, content: str):
        """
        Add a message to the memory buffer.
        """
        pass

    def add_user_message(self, content: str):
        """
        Add a user message to the memory buffer.
        """
        self.add_message(role="user", content=content)

    def add_assistant_message(self, content: str):
        """
        Add an assistant message to the memory buffer.
        """
        self.add_message(role="assistant", content=content)

    @abstractmethod
    def set_system_message(self, content: str):
        """
        Add a system message to the memory buffer.
        """
        pass

    @abstractmethod
    def get_messages(self):
        """
        Get the messages from the memory buffer.
        """
        pass

    @abstractmethod
    def get_system_message(self):
        """
        Get the system message from the memory buffer.
        """
        pass

    ###########
    # Helpers #
    ###########

    def to_messages_template(self):
        """
        Convenience method to convert the memory to a MessagesTemplate object.
        """
        template = MessagesTemplate(
            source=Messages(
                messages=self.get_messages(), system=self.get_system_message()
            )
        )
        return template


class FleetingMemory(BaseMemory):
    """
    A simple memory context for the agent, only kept in memory (lol).
    """

    def __init__(self):
        """Initialize the memory object."""
        self.system = None
        self.messages = []

    def add_message(self, role: str, content: str):

        if role == "system":
            self.system = content
        elif role in ["user", "assistant"]:
            self.messages.append(Message(role=role, content=content))
        else:
            raise ValueError(
                f"Role must be 'system', 'user', or 'assistant'. Got {role}"
            )

    def set_system_message(self, content: str):
        self.add_message(role="system", content=content)

    def add_messages(self, messages: list[Message]):
        self.messages.extend(messages)

    def get_messages(self):
        return self.messages

    def get_system_message(self):
        return self.system if hasattr(self, "system") else None


if __name__ == "__main__":

    # Create memory object
    memory = FleetingMemory()

    # Add a system message
    memory.set_system_message("This is a system message")

    # Add a user message
    memory.add_user_message("This is a user message")

    # Add an assistant message
    memory.add_assistant_message("This is an assistant message")

    # Print the messages
    logger.info(
        f"System message <{type(memory.get_system_message())}>: {memory.get_system_message()}"
    )
    logger.info(f"Messages <{type(memory.get_messages())}>: {memory.get_messages()}")



"""
Base Agent class
"""

import logging
from abc import abstractmethod

from fence import LLM
from fence.links import logger as link_logger

logger = logging.getLogger(__name__)

# Suppress the link logger
link_logger.setLevel("INFO")


class BaseAgent:
    """Base Agent class"""

    def __init__(
        self,
        identifier: str | None = None,
        model: LLM = None,
        description: str | None = None,
        environment: dict | None = None,
    ):
        """
        Initialize the Agent object.

        :param identifier: An identifier for the agent. If none is provided, the class name will be used.
        :param model: An LLM model object.
        :param description: A description of the agent.
        :param environment: A dictionary of environment variables to pass to delegates and tools.
        """
        self.environment = environment or {}
        self.identifier = identifier or self.__class__.__name__
        self.model = model
        self.description = description

    @abstractmethod
    def run(self, prompt: str) -> str:
        """
        Run the agent

        :param prompt: The initial prompt to feed to the LLM
        """
        raise NotImplementedError

    def format_toml(self):
        """
        Returns a TOML-formatted key-value pair of the agent name,
        the description (docstring) of the agent.
        """

        # Preformat the arguments
        toml_string = f"""[[agents]]
agent_name = "{self.identifier}"
agent_description = "{self.description or self.__doc__}"
"""

        return toml_string



"""
Logging utils
"""

import logging
import logging.config
from enum import Enum

DEFAULT_LOGGING_FORMAT = "[%(levelname)s][%(name)s.%(funcName)s:%(lineno)d] %(message)s"

#####################
# Base configuration #
#####################


class ColorFormatter(logging.Formatter):
    """
    A custom formatter that adds joy to the log messages.
    """

    COLORS = {
        "DEBUG": "\033[1;34m",  # Bold blue
        "INFO": "\033[1;32m",  # Bold green
        "WARNING": "\033[1;33m",  # Bold yellow
        "ERROR": "\033[1;31m",  # Bold red
        "CRITICAL": "\033[1;35m",  # Bold magenta
    }
    EMOJIS = {
        "DEBUG": "🔵",
        "INFO": "ℹ️",
        "WARNING": "😬",
        "ERROR": "❌",
        "CRITICAL": "🚨",
    }

    def format(self, record):
        level_color = self.COLORS.get(record.levelname, "\033[0m")  # Default to reset
        reset_color = "\033[0m"

        # Format the message with the LOGGING_FORMAT
        formatted_message = super().format(record)

        # Add timestamp, level, and color to the message
        time = self.formatTime(record, "%Y-%m-%d %H:%M:%S")
        formatted_message = f"[{time}] [{record.levelname}] {self.EMOJIS.get(record.levelname, '')} [{record.module}.{record.funcName}:{record.lineno}] {formatted_message}"

        return f"{level_color}{formatted_message}{reset_color}"


class LogLevels(Enum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


def setup_logging(name: str = "root", log_level: str = None, serious_mode: bool = None):
    """
    Setup logging for use in applications.
    :param name: name of the logger
    :param log_level: log level as a string (DEBUG, INFO, WARNING, ERROR, CRITICAL)
    :param serious_mode: whether to use the serious mode
    :return: logger instance
    """

    # Base configuration
    config_dict = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "simple": {
                "format": "%(asctime)s [%(levelname)s] [%(name)s.%(funcName)s:%(lineno)d] %(message)s"
            },
            "color": {"()": ColorFormatter},
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": "INFO",
                "formatter": "simple",
                "stream": "ext://sys.stdout",
            }
        },
        "root": {"level": "INFO", "handlers": ["console"]},
    }

    # Set the log level if it is not provided
    if log_level is not None:
        if log_level.upper() not in LogLevels.__members__:
            raise ValueError(
                f"Invalid log level: {log_level}. Should be one of {', '.join(LogLevels.__members__)}."
            )
        config_dict["root"]["level"] = log_level.upper()
        config_dict["handlers"]["console"]["level"] = log_level.upper()

    # Set the serious mode if it is not provided
    if serious_mode is not None:
        config_dict["handlers"]["console"]["formatter"] = (
            "simple" if serious_mode else "color"
        )

    # Configure the logging module with the config file
    logging.config.dictConfig(config_dict)

    # Suppress logging from boto3 and botocore
    logging.getLogger("boto3").setLevel(logging.CRITICAL)
    logging.getLogger("botocore").setLevel(logging.CRITICAL)
    logging.getLogger("urllib3").setLevel(logging.CRITICAL)

    # Get the logger
    logger = logging.getLogger(name)

    return logger


if __name__ == "__main__":

    logger = setup_logging(log_level="debug", serious_mode=False)
    logger.debug("This is a debug message")
    logger.info("This is an info message")
    logger.warning("This is a warning message")
    logger.error("This is an error message")
    logger.critical("This is a critical message")
